class Attraction2D implements Force
{
  Particle a,b;
  float k;
  float distanceMin;
  float distanceMinSquared;
  boolean on = true;

  Attraction2D( Particle a, Particle b, float k, float distanceMin )
  {
    this.a = a;
    this.b = b;
    this.k = k;
    this.distanceMin = distanceMin;
    this.distanceMinSquared = distanceMin*distanceMin;
  }

  void setMinimumDistance( float d )
  {
    distanceMin = d;
    distanceMinSquared = d*d;
  }

  void apply()
  {
    if ( on && ( a.isFree() || b.isFree() ) )
    {
      float a2bX = a.position().x() - b.position().x();
      float a2bY = a.position().y() - b.position().y();
      
      float a2bDistanceSquared = a2bX*a2bX + a2bY*a2bY;

      if ( a2bDistanceSquared < distanceMinSquared )
        a2bDistanceSquared = distanceMinSquared;

      float force = k / a2bDistanceSquared;

      //Add this to make the force drop off with 1/sq(d) instead of 1/d
      float length = (float)Math.sqrt( a2bDistanceSquared );
      
      a2bX /= length;
      a2bY /= length;

      // multiply by force 

      a2bX *= force;
      a2bY *= force;

      // apply
      if ( a.isFree() )
        a.force().add( -a2bX,-a2bY, 0 );
      if ( b.isFree() )
        b.force().add( a2bX, a2bY, 0 );
    }
  }
  
  float strength() {
    return k;
  }
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
}

class SpiralForce implements Force
{
  Particle a,b;
  float k;
  float distanceMin;
  float distanceMinSquared;
  boolean on = true;
  
  SpiralForce( Particle a, Particle b, float k, float distanceMin )
  {
    this.a = a;
    this.b = b;
    this.k = k;
    this.distanceMin = distanceMin;
    this.distanceMinSquared = distanceMin*distanceMin;
  }

  void setMinimumDistance( float d )
  {
    distanceMin = d;
    distanceMinSquared = d*d;
  }

  void apply()
  {
    if ( on && ( a.isFree() || b.isFree() ) )
    {
      float a2bX = a.position().x() - b.position().x();
      float a2bY = a.position().y() - b.position().y();
      
      float a2bDistanceSquared = a2bX*a2bX + a2bY*a2bY;

      if ( a2bDistanceSquared < distanceMinSquared )
        a2bDistanceSquared = distanceMinSquared;

      float force = k / a2bDistanceSquared;

      //Add this to make the force drop off with 1/sq(d) instead of 1/d
      //float length = (float)Math.sqrt( a2bDistanceSquared );
      
      //a2bX /= length;
      //a2bY /= length;

      // multiply by force 

      a2bX *= force;
      a2bY *= force;

      // apply
      //The only difference between spiral and a regular force is we rotate x and y when we apply it
      if ( a.isFree() )
        a.force().add( -a2bY,a2bX, 0 );
      if ( b.isFree() )
        b.force().add( a2bY, -a2bX, 0 );
    }
  }
  
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
}

class VariableAttraction implements Force
{
  Particle a,b;
  float k;
  float distanceMin;
  float distanceMinSquared;
  float exponent;
  boolean on = true;

  VariableAttraction( Particle a, Particle b, float k, float distanceMin )
  {
    this.a = a;
    this.b = b;
    this.k = k;
    this.distanceMin = distanceMin;
    this.distanceMinSquared = distanceMin*distanceMin;
    exponent = 2;
  }

  void setMinimumDistance( float d )
  {
    distanceMin = d;
    distanceMinSquared = d*d;
  }

  void apply()
  {
    if ( on && ( a.isFree() || b.isFree() ) )
    {
      float a2bX = a.position().x() - b.position().x();
      float a2bY = a.position().y() - b.position().y();
      float a2bZ = a.position().z() - b.position().z();
      
      float a2bDistanceSquared = a2bX*a2bX + a2bY*a2bY + a2bZ;

      if ( a2bDistanceSquared < distanceMinSquared )
        a2bDistanceSquared = distanceMinSquared;

      float length = (float)Math.sqrt( a2bDistanceSquared );

      float force = k / pow(length,exponent);
      
      a2bX /= length;
      a2bY /= length;
      a2bZ /= length;

      // multiply by force 

      a2bX *= force;
      a2bY *= force;

      // apply
      if ( a.isFree() )
        a.force().add( -a2bX,-a2bY, -a2bZ );
      if ( b.isFree() )
        b.force().add( a2bX, a2bY, a2bZ );
    }
  }
  
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
  
}

class NoiseForce implements Force {
  Particle p;
  float noiseScale;
  float k;
  boolean on = true;
  
  NoiseForce(Particle p, float k, float noiseScale) {
    this.p = p;
    this.k = k;
    this.noiseScale = noiseScale;
  }
  
  void apply() {
    if(on && p.isFree()) {
      float forceX = k*(noise(p.position().x()*noiseScale,p.position().y()*noiseScale,0)-.5);
      float forceY = k*(noise(p.position().x()*noiseScale,p.position().y()*noiseScale,50)-.5);
      p.force().add(forceX,forceY,0);
    }
  }
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
  
}

class NoisePotential2D implements Force {
  Particle p;
  float noiseScale;
  float k;
  float DELTA = 1;
  boolean on = true;
  
  NoisePotential2D(Particle p, float k, float noiseScale) {
    this.p = p;
    this.k = k;
    this.noiseScale = noiseScale;
  }
  
  void apply() {
    if(on && p.isFree()) {
      float potentialX0 = noise((p.position().x()-DELTA)*noiseScale,p.position().y()*noiseScale);
      float potentialX1 = noise((p.position().x()+DELTA)*noiseScale,p.position().y()*noiseScale);
      float potentialY0 = noise(p.position().x()*noiseScale,(p.position().y()-DELTA)*noiseScale);
      float potentialY1 = noise(p.position().x()*noiseScale,(p.position().y()+DELTA)*noiseScale);
      
      p.force().add(k*(potentialX1-potentialX0)/(2*DELTA),k*(potentialY1-potentialY0)/(2*DELTA),0);
    }
  }
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
}

class ImagePotential implements Force {
  Particle p;
  float k;
  PImage image;
  boolean on = true;
  
  ImagePotential(Particle p, float k, PImage image) {
    this.p = p;
    this.k = k;
    this.image = image;
  }
  
  void apply() {
    if(on && p.isFree()) {
      int xpos = int(p.position().x());
      int ypos = int(p.position().y());
      if(xpos>=image.width || xpos < 0 || ypos >= image.height || ypos < 0) return;
      float potentialX0,potentialX1,potentialY0,potentialY1;
      potentialX0=potentialX1=potentialY0=potentialY1=255;
      if(xpos-1 > 0) potentialX0 = brightness(image.pixels[ypos*image.width+xpos-1]); 
      if(xpos+1 < image.width) potentialX1 = brightness(image.pixels[ypos*image.width+xpos+1]); 
      if(ypos-1 > 0) potentialY0 = brightness(image.pixels[(ypos-1)*image.width+xpos]); 
      if(ypos+1 < image.height) potentialY1 = brightness(image.pixels[(ypos+1)*image.width+xpos]); 
      
      p.force().add(k*(potentialX1-potentialX0)/2.0,k*(potentialY1-potentialY0)/2.0,0);
    }
  }
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
}

class RandomForce2D implements Force {
  Particle p;
  float k;
  boolean on = true;
  
  RandomForce2D(Particle p, float k) {
    this.p = p;
    this.k = k;
  }
  
  void apply() {
    if(on && p.isFree()) {
      p.force().add(random(-k,k),random(-k,k),0);
    }
  }
  boolean isOn() {return on;}
  boolean isOff() {return !on;}
  void turnOff() {on = false;}
  void turnOn() {on = true;}
}
